/*
 * Copyright (c) 1997 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 * 
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This file is part of the Blizzard source distribution
 * originally written by Babak Falsafi, Alvin Lebeck, Steven
 * K. Reinhardt, and Yannis Schoinas and subsequently maintained 
 * by Babak Falsafi, Chris Lukas, Steven K. Reinhardt, Yannis Schoinas
 * and Brian Toonen.
 */
/*
 * alloc_struct.h
 *
 * Fast, low-overhead allocation for fixed-size structures.
 * Save space & time by assuming the user knows the size of the
 * structure being freed.  Can be made transparent by using sizeof().
 *
 * July 24, 1991 Steve Reinhardt
 */
#ifndef _alloc_struct_h
#define _alloc_struct_h

#include "../master/tppi_vm.h"	/* for TPPI_PAGE_{SIZE,SHIFT} */
#include "../master/fatal_error.h"
#include "../master/local_lock.h"

#ifndef DIV_CEIL
#define DIV_CEIL(n,d)   ( ((n)+(d)-1)/(d) )
#endif

#define MAX_ALLOC_SIZE	(TPPI_PAGE_SIZE)

#define ALLOC_QUANTUM	sizeof(double)

#define MAX_BUCKET	DIV_CEIL(MAX_ALLOC_SIZE,ALLOC_QUANTUM)

extern atomic_sequence_t alloc_struct_seq;

/*
 * Allocate a structure of size sz.  Call alloc_more_structs()
 * if there aren't any on free list.  This function should
 * not be called directly; use alloc_struct() macro.
 */
INLINE 
char *alloc_struct_f(unsigned sz)
{
    extern char *struct_free[];
    extern char *_NI_alloc_more_structs(int);
    int b;
    char *p;

    start_atomic_sequence(&alloc_struct_seq);

    if (sz > MAX_ALLOC_SIZE)
	fatal_error("alloc_struct size (%d) > MAX_ALLOC_SIZE (%d)\n",
		    sz, MAX_ALLOC_SIZE);

    b = DIV_CEIL(sz, ALLOC_QUANTUM);
    p = struct_free[b];
    
    if (p)
	struct_free[b] = *(char **)p;
    else
	p = _NI_alloc_more_structs(b);
    
    end_atomic_sequence(&alloc_struct_seq);

    return p;
}

/*
 * Free structure of size 'sz', which must correspond to the size
 * provided in the call to alloc_struct()_f().
 * Do not call directly; use free_struct() macro.
 */
INLINE 
void free_struct_f(char *p, unsigned sz)
{
    int b;
    extern char *struct_free[];

    start_atomic_sequence(&alloc_struct_seq);
    
    b = DIV_CEIL(sz, ALLOC_QUANTUM);
    *(char **)p = struct_free[b];
    struct_free[b] = p;

    end_atomic_sequence(&alloc_struct_seq);
}

#ifdef MALLOC_DEBUG

#define alloc_struct(t) malloc(sizeof(t)) 
#define free_struct(p)	free(p) 
#define alloc_bytes(n)	malloc(n) 
#define free_bytes(p,n)	free(p)

#else

#define alloc_struct(t) ((t *) alloc_struct_f(sizeof(t)))
#define free_struct(p)	(free_struct_f((char *)p, sizeof(*p)))
#define alloc_bytes(n)	((void *) alloc_struct_f(n))
#define free_bytes(p,n)	(free_struct_f((char *)p, n))

#endif MALLOC_DEBUG

#endif
