/*
 * Copyright (c) 1997 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 * 
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This file is part of the Blizzard source distribution
 * originally written by Babak Falsafi, Alvin Lebeck, Steven
 * K. Reinhardt, and Yannis Schoinas and subsequently maintained 
 * by Babak Falsafi, Chris Lukas, Steven K. Reinhardt, Yannis Schoinas
 * and Brian Toonen.
 */
#ifndef _mcs_lock_h
#define _mcs_lock_h

#include <assert.h>

/*
 * Basic MCS lock structure.
 */
typedef struct MCS_qnode_tag 
{
    struct MCS_qnode_tag * volatile next;
    volatile int locked;
} MCS_qnode_t;

/*
 * An "MCS lock" is just a pointer to a qnode (NULL if it's free,
 * otherwise pointing to the tail of the queue).  Since this is
 * where nodes competitively swap to get the lock, it's volatile
 * (note that the ptr itself is volatile, not what it points to).
 */
typedef MCS_qnode_t * volatile MCS_lock_t;

/*
 * MCS acquire/release calls.  Caller must provide a local qnode struct.
 * The same ptr must be passed to release as was passed in on the
 * corresponding acquire.  This qnode should be in locally-allocated
 * shared memory.
 */
void MCS_acquire(MCS_lock_t *, MCS_qnode_t *);
void MCS_release(MCS_lock_t *, MCS_qnode_t *);

/*
 * For convenience, we provide lock() and unlock() where
 * the allocation of a local qnode is hidden from the user.
 * This is done by keeping a ptr to the local qnode of the locker
 * (the head of the queue) in a struct along with the lock itself
 * (the tail of the queue).
 */
typedef struct Master_Lock *LockPtr;


struct Master_Lock 
{
    MCS_lock_t	  master;	/* actual MCS lock */
    MCS_qnode_t  *local_qnode;	/* ptr to head of queue; used only by locker */
    int pad[6];   /* Allocate a full cache block to prevent false sharing */
};

/*
 * Declare a lock.
 * This pointer can be in private or shared memory as long as only one node
 * does the MCS_LOCKINIT(lockname) and all nodes get the value from that.
 */
#define MCS_LOCKDEC(lockname)  LockPtr lockname;

/* Initialize a lock */
#define MCS_LOCKINIT(lockname)						      \
do {									      \
	lockname = (LockPtr)shm_malloc(sizeof(struct Master_Lock));	      \
	assert(lockname != 0);						      \
	lockname->master = 0;						      \
	lockname->local_qnode = 0;					      \
} while (0);

/* Initialize a lock, using sbrk instead of malloc.  This is for malloc. */
#define MCS_SBRK_LOCKINIT(lockname)					      \
do {									      \
	lockname = (LockPtr)shm_sbrk(sizeof(struct Master_Lock));	      \
	lockname->master = 0;						      \
	lockname->local_qnode = 0; 					      \
} while (0)

void mcs_lock(LockPtr);
void mcs_unlock(LockPtr);

#endif _mcs_lock_h
