/*
 * Copyright (c) 1997 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 * 
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This file is part of the Blizzard source distribution
 * originally written by Babak Falsafi, Alvin Lebeck, Steven
 * K. Reinhardt, and Yannis Schoinas and subsequently maintained 
 * by Babak Falsafi, Chris Lukas, Steven K. Reinhardt, Yannis Schoinas
 * and Brian Toonen.
 */
/*
 * utils_inlines.h
 *
 * This file provides common definitions and declarations that are
 * valid for both the CP and the PN's.
 */

#ifndef _tppi_util_inlines_h
#define _tppi_util_inlines_h

#include <memory.h>
#include "../master/types.h"
#include "../master/fatal_error.h"

#define copy32 _NI_copy32
#define copy8 _NI_copy8
#define copy4 _NI_copy4
#define fcopy32 _NI_fcopy32

#ifdef __cplusplus
extern "C" {
#endif

#define cbarrier() asm volatile("")

#define xmalloc _NI_xmalloc
/*
 * Call malloc, panic if it fails.
 */
INLINE
void *xmalloc(unsigned int nbytes)
{
    extern void *malloc(unsigned int);
    void *p;

    if ((p = malloc(nbytes)) == NULL)
    {
	fatal_error("Out of memory!\n");
    }
    return p;
}

/*
 * Call xmalloc, then zero allocated memory.
 */
INLINE
void *xmalloc_z(unsigned int nbytes)
{
    void *p = xmalloc(nbytes);
    memset(p, 0, nbytes);
    return p;
}

INLINE 
Uint32 swap(volatile Uint32 *ptr, Uint32 val)
{
    Uint32 retval = val;

    asm volatile ("swap %1,%0" : "=r" (retval), "=m" (*ptr) : "0" (retval));

    return retval;
}

/* ldstub is a byte-wide swap that always writes 0xff */
INLINE 
Uint8 ldstub(volatile Uint8 *ptr)
{
    Uint8 retval;

    asm volatile ("ldstub %1,%0" : "=r" (retval), "=m" (*ptr) : );

    return retval;
}

INLINE 
void set32(void *va, int val, int size)
{
    register Uint64 tmp1,tmp2;

    tmp2 = hiLo32To64(val, val);

    asm volatile ("\n
0:     std %2, [%1]\n
       std %2, [%1 + 0x8]\n
       std %2, [%1 + 0x10]\n
       subcc %0,32,%0\n
       bz  1f\n
       std %2, [%1 + 0x18]\n
       ba 0b\n
       add %1,32,%1\n
1:" : "=r" (size), "=r" (va), "=r" (tmp1)
    : "0" (size), "1" (va), "2" (tmp2));
}

INLINE 
void copy32(void *from, void *to, int size)
{
    register Uint64 tmp1,tmp2,tmp3,tmp4;

    asm volatile ("\n
0:     ldd [%1],%3\n
       ldd [%1 + 0x8],%4\n
       ldd [%1 + 0x10],%5\n
       ldd [%1 + 0x18],%6\n
       std %3,[%2]\n
       std %4,[%2 + 0x8]\n
       std %5,[%2 + 0x10]\n
       subcc %0,32,%0\n
       bz  1f\n
       std %6,[%2 + 0x18]\n
       add %1,32,%1\n
       ba 0b\n
       add %2,32,%2\n
1:" : "=r" (size), "=r" (from), "=r" (to),
      "=r" (tmp1), "=r" (tmp2), "=r" (tmp3), "=r" (tmp4)
    : "0" (size), "1" (from), "2" (to));
}

INLINE 
void copy4(void *from, void *to, int size)
{
    register int i, *tmp1 = (int *)from, *tmp2 = (int *)to;

    for (i = 0; i < size/4; ++i)
    {
	tmp2[i] = tmp1[i];
    }
}

INLINE
void copy8(void *from, void *to, int size)
{
    register int i;
    register Uint64 *tmp1 = (Uint64 *)from, *tmp2 = (Uint64 *)to;

    for (i = 0; i < size/8; i++)
    {
	tmp2[i] = tmp1[i];
    }
}

INLINE
void fcopy32(void *from, void *to, int size)
{
   if (size == 0)
     return;
   if (((int)from & 0x7) == 0 && ((int)to & 0x7) == 0)
   {
       if (size % 32 == 0)
       {
	   copy32(from, to, size);
       }
       else if (size % 8 == 0)
       {
	   copy8(from, to, size);
       }
       else
       {
	   copy4(from, to, size);
       }
   }
   else
   {
       copy4(from, to , size);
   }
}

INLINE 
void dump_fps(void *buf)
{
    asm("std %%f0,[%0]" : : "r" (buf));
    asm("std %%f2,[%0 + 0x8]" : : "r" (buf));
    asm("std %%f4,[%0 + 0x10]" : : "r" (buf));
    asm("std %%f6,[%0 + 0x18]" : : "r" (buf));
    asm("std %%f8,[%0 + 0x20]" : : "r" (buf));
    asm("std %%f10,[%0 + 0x28]" : : "r" (buf));
    asm("std %%f12,[%0 + 0x30]" : : "r" (buf));
    asm("std %%f14,[%0 + 0x38]" : : "r" (buf));
    asm("std %%f16,[%0 + 0x40]" : : "r" (buf));
    asm("std %%f18,[%0 + 0x48]" : : "r" (buf));
    asm("std %%f20,[%0 + 0x50]" : : "r" (buf));
    asm("std %%f22,[%0 + 0x58]" : : "r" (buf));
    asm("std %%f24,[%0 + 0x60]" : : "r" (buf));
    asm("std %%f26,[%0 + 0x68]" : : "r" (buf));
    asm("std %%f28,[%0 + 0x70]" : : "r" (buf));
    asm("std %%f30,[%0 + 0x78]" : : "r" (buf));
}

INLINE 
void load_fps(void *buf)
{
    asm("ldd [%0],%%f0" : : "r" (buf));
    asm("ldd [%0 + 0x8],%%f2 " : : "r" (buf));
    asm("ldd [%0 + 0x10],%%f4 " : : "r" (buf));
    asm("ldd [%0 + 0x18],%%f6 " : : "r" (buf));
    asm("ldd [%0 + 0x20],%%f8 " : : "r" (buf));
    asm("ldd [%0 + 0x28],%%f10" : : "r" (buf));
    asm("ldd [%0 + 0x30],%%f12" : : "r" (buf));
    asm("ldd [%0 + 0x38],%%f14" : : "r" (buf));
    asm("ldd [%0 + 0x40],%%f16" : : "r" (buf));
    asm("ldd [%0 + 0x48],%%f18" : : "r" (buf));
    asm("ldd [%0 + 0x50],%%f20" : : "r" (buf));
    asm("ldd [%0 + 0x58],%%f22" : : "r" (buf));
    asm("ldd [%0 + 0x60],%%f24" : : "r" (buf));
    asm("ldd [%0 + 0x68],%%f26" : : "r" (buf));
    asm("ldd [%0 + 0x70],%%f28" : : "r" (buf));
    asm("ldd [%0 + 0x78],%%f30" : : "r" (buf));
}

/*
 * ceil(n/d)
 */
INLINE 
int divCeil(int n, int d)
{
    return (n + d - 1) / d;
}

/*
 * floor(log2(n))... works by finding position of MSB set.
 * Returns -1 if n == 0.
 */
INLINE 
int floorLog2(unsigned n)
{
    int p = 0;

    if (n == 0)
	return -1;

    if (n & 0xffff0000) { p += 16; n >>= 16; }
    if (n & 0x0000ff00)	{ p +=  8; n >>=  8; }
    if (n & 0x000000f0) { p +=  4; n >>=  4; }
    if (n & 0x0000000c) { p +=  2; n >>=  2; }
    if (n & 0x00000002) { p +=  1; }

    return p;
}

/*
 * log base 2 of argument... actually returns floor(log2(n)).
 */
INLINE 
int ilog2(unsigned n)
{
    return floorLog2(n);
}

INLINE 
int ceilLog2(unsigned n)
{
    return floorLog2(n-1)+1;
}

INLINE 
int isPowerOf2(unsigned n)
{
    return ceilLog2(n) == floorLog2(n);
}

#ifdef __cplusplus
};
#endif

#endif _tppi_util_inlines_h
